from unittest.mock import Mock

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection import Connection, ConnectParams, MessageHandlers
from ohmqtt.packet import (
    MQTTConnAckPacket,
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
)
from ohmqtt.property import MQTTConnAckProps, MQTTPublishProps
from ohmqtt.session import Session
from ohmqtt.subscriptions import Subscriptions
from ohmqtt.topic_alias import AliasPolicy


@pytest.fixture
def mock_handlers(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=MessageHandlers)  # type: ignore[no-any-return]


@pytest.fixture
def mock_connection(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=Connection)  # type: ignore[no-any-return]


@pytest.fixture
def mock_subscriptions(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=Subscriptions)  # type: ignore[no-any-return]


def test_session_publish_qos0(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    mock_connection.can_send.return_value = True

    handle = session.publish("test/topic", b"test payload")
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
    ))
    mock_connection.send.reset_mock()
    assert handle.is_acked() is False
    assert handle.wait_for_ack() is False


def test_session_publish_qos1(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    session.server_receive_maximum = 20
    mock_connection.can_send.return_value = True

    handle = session.publish("test/topic", b"test payload", qos=1)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        qos=1,
        packet_id=1,
    ))
    mock_connection.send.reset_mock()
    assert handle.is_acked() is False
    assert handle.wait_for_ack(0.001) is False

    session.handle_puback(MQTTPubAckPacket(packet_id=1))
    assert handle.is_acked() is True
    assert handle.wait_for_ack(0.001) is True


def test_session_publish_qos2(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    session.server_receive_maximum = 20
    mock_connection.can_send.return_value = True

    handle = session.publish("test/topic", b"test payload", qos=2)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        qos=2,
        packet_id=1,
    ))
    mock_connection.send.reset_mock()
    assert handle.is_acked() is False
    assert handle.wait_for_ack(0.001) is False

    session.handle_pubrec(MQTTPubRecPacket(packet_id=1))
    assert handle.is_acked() is False
    assert handle.wait_for_ack(0.001) is False

    mock_connection.send.assert_called_with(MQTTPubRelPacket(packet_id=1))
    mock_connection.send.reset_mock()

    session.handle_pubcomp(MQTTPubCompPacket(packet_id=1))
    assert handle.is_acked() is True
    assert handle.wait_for_ack(0.001) is True


@pytest.mark.parametrize("db_path", [":memory:", ""])
@pytest.mark.parametrize("qos", [0, 1, 2])
def test_session_publish_alias(db_path: str, qos: int, mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection, db_path=db_path)
    session.set_params(ConnectParams(client_id="test_client", clean_start=True))
    mock_connection.can_send.return_value = True

    session.handle_connack(MQTTConnAckPacket(properties=MQTTConnAckProps(TopicAliasMaximum=255)))

    session.publish("test/topic1", b"test payload", qos=qos, alias_policy=AliasPolicy.NEVER)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic1",
        payload=b"test payload",
        qos=qos,
        packet_id=1 if qos > 0 else 0,
    ))
    mock_connection.send.reset_mock()

    session.publish("test/topic2", b"test payload", qos=qos, alias_policy=AliasPolicy.TRY)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic2",
        payload=b"test payload",
        qos=qos,
        packet_id=2 if qos > 0 else 0,
        properties=MQTTPublishProps(TopicAlias=1),
    ))
    mock_connection.send.reset_mock()

    if qos > 0:
        with pytest.raises(ValueError):
            session.publish("test/topic3", b"test payload", qos=qos, alias_policy=AliasPolicy.ALWAYS)
    else:
        session.publish("test/topic3", b"test payload", qos=qos, alias_policy=AliasPolicy.ALWAYS)
        mock_connection.send.assert_called_with(MQTTPublishPacket(
            topic="test/topic3",
            payload=b"test payload",
            qos=qos,
            packet_id=3 if qos > 0 else 0,
            properties=MQTTPublishProps(TopicAlias=2),
        ))
        mock_connection.send.reset_mock()


def test_session_handle_publish_qos0(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    mock_connection.can_send.return_value = True

    packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        qos=0,
    )
    session.handle_publish(packet)
    mock_subscriptions.handle_publish.assert_called_once_with(packet)


def test_session_handle_publish_qos1(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    mock_connection.can_send.return_value = True

    packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        packet_id=3,
        qos=1,
    )
    session.handle_publish(packet)
    mock_subscriptions.handle_publish.assert_called_once_with(packet)
    mock_connection.send.assert_called_once_with(MQTTPubAckPacket(packet_id=3))


def test_session_handle_publish_qos2(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    mock_connection.can_send.return_value = True

    packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        packet_id=3,
        qos=2,
    )
    session.handle_publish(packet)
    mock_subscriptions.handle_publish.assert_called_once_with(packet)
    mock_subscriptions.handle_publish.reset_mock()
    mock_connection.send.assert_called_once_with(MQTTPubRecPacket(packet_id=3))
    mock_connection.send.reset_mock()

    # Filter duplicates.
    session.handle_publish(packet)
    mock_subscriptions.handle_publish.assert_not_called()
    mock_connection.send.assert_called_once_with(MQTTPubRecPacket(packet_id=3))
    mock_connection.send.reset_mock()

    session.handle_pubrel(MQTTPubRelPacket(packet_id=3))
    mock_connection.send.assert_called_once_with(MQTTPubCompPacket(packet_id=3))
    mock_connection.send.reset_mock()
    # After PUBREL, message with same packet_id should be treated as a new application message.
    session.handle_publish(packet)
    mock_subscriptions.handle_publish.assert_called_once_with(packet)
    mock_connection.send.assert_called_once_with(MQTTPubRecPacket(packet_id=3))


def test_session_slots(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    assert not hasattr(session, "__dict__")
    assert all(hasattr(session, attr) for attr in session.__slots__), \
        [attr for attr in session.__slots__ if not hasattr(session, attr)]



from unittest.mock import Mock, MagicMock
import weakref

import pytest
from pytest_mock import MockerFixture

from ohmqtt.client import Client
from ohmqtt.connection import Connection, MessageHandlers, InvalidStateError
from ohmqtt.mqtt_spec import MQTTReasonCode
from ohmqtt.packet import (
    MQTTPublishPacket,
    MQTTSubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubscribePacket,
    MQTTUnsubAckPacket,
    MQTTConnAckPacket,
)
from ohmqtt.property import MQTTSubscribeProps, MQTTUnsubscribeProps
from ohmqtt.subscriptions import Subscriptions, RetainPolicy


def dummy_callback(client: Client, packet: MQTTPublishPacket) -> None:
    """Dummy SubscribeCallback for testing."""


@pytest.fixture
def mock_client(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=Client)  # type: ignore[no-any-return]


@pytest.fixture
def mock_connection(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=Connection)  # type: ignore[no-any-return]


@pytest.fixture
def mock_handlers(mocker: MockerFixture) -> MagicMock:
    return mocker.MagicMock(spec=MessageHandlers)  # type: ignore[no-any-return]


def test_subscriptions_registration(mock_client: Mock, mock_connection: Mock) -> None:
    handlers = MessageHandlers()
    with handlers:
        subscriptions = Subscriptions(handlers, mock_connection, weakref.ref(mock_client))
    assert subscriptions.handle_suback in handlers.get_handlers(MQTTSubAckPacket)
    assert subscriptions.handle_unsuback in handlers.get_handlers(MQTTUnsubAckPacket)
    assert subscriptions.handle_connack in handlers.get_handlers(MQTTConnAckPacket)


@pytest.mark.parametrize("max_qos", [0, 1, 2])
@pytest.mark.parametrize("no_local", [True, False])
@pytest.mark.parametrize("retain_as_published", [True, False])
@pytest.mark.parametrize("retain_policy", [RetainPolicy.NEVER, RetainPolicy.ONCE, RetainPolicy.ALWAYS])
def test_subscriptions_subscribe_opts(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
    max_qos: int,
    no_local: bool,
    retain_as_published: bool,
    retain_policy: RetainPolicy
) -> None:
    """Test subscribing with options."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe(
        "test/topic",
        dummy_callback,
        max_qos=max_qos,
        share_name="test_share",
        no_local=no_local,
        retain_as_published=retain_as_published,
        retain_policy=retain_policy,
        sub_id=23,
        user_properties=[("key", "value")],
    )
    expected_opts = max_qos | (retain_policy << 4) | (retain_as_published << 3) | (no_local << 2)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("$share/test_share/test/topic", expected_opts)],
        packet_id=1,
        properties=MQTTSubscribeProps(
            SubscriptionIdentifier={23},
            UserProperty=(("key", "value"),),
        ),
    ))


def test_subscriptions_handle_unsubscribe(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test using a subscription handle to unsubscribe."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    sub_handle = subscriptions.subscribe("test/topic", dummy_callback)
    assert sub_handle is not None
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    ))

    unsub_handle = sub_handle.unsubscribe()
    assert unsub_handle is not None
    mock_connection.send.assert_called_with(MQTTUnsubscribePacket(
        topics=["test/topic"],
        packet_id=1,
        properties=MQTTUnsubscribeProps(),
    ))

    unsuback_packet = MQTTUnsubAckPacket(packet_id=1)
    subscriptions.handle_unsuback(unsuback_packet)

    assert unsub_handle.wait_for_ack(timeout=0.1) == unsuback_packet
    assert unsub_handle.ack == unsuback_packet


def test_subscriptions_wait_for_suback(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test waiting for a SUBACK packet."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    handle = subscriptions.subscribe("test/topic", dummy_callback)
    assert handle is not None

    # Simulate receiving a SUBACK packet
    suback_packet = MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    )
    subscriptions.handle_suback(suback_packet)

    assert handle.wait_for_ack(timeout=0.1) == suback_packet
    assert handle.ack == suback_packet


@pytest.mark.parametrize("session_present", [True, False])
def test_subscriptions_subscribe_failure(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
    session_present: bool
) -> None:
    """Test replaying SUBSCRIBE packets on reconnection after calling in a bad state."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    mock_connection.send.side_effect = InvalidStateError("TEST")
    handle = subscriptions.subscribe("test/topic", dummy_callback)

    assert handle is None

    mock_connection.send.side_effect = None
    mock_connection.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(session_present=session_present))

    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    ))
    mock_connection.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(session_present=session_present))

    if not session_present:
        # If the session was not present, we should send the SUBSCRIBE packet again.
        mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
            topics=[("test/topic", 2)],
            packet_id=1,
        ))


@pytest.mark.parametrize("session_present", [True, False])
def test_subscriptions_multi_subscribe(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
    session_present: bool
) -> None:
    """Test subscribing to the same topic multiple times."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    handle1 = subscriptions.subscribe("test/topic", dummy_callback)
    handle2 = subscriptions.subscribe("test/topic", dummy_callback)

    assert handle1 is not None
    assert handle2 is not None
    assert handle1 != handle2

    assert mock_connection.send.call_count == 2
    assert mock_connection.send.call_args_list[0][0][0] == MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    )
    assert mock_connection.send.call_args_list[1][0][0] == MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=2,
    )
    mock_connection.reset_mock()

    suback_packet = MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    )
    subscriptions.handle_suback(suback_packet)

    assert handle1.wait_for_ack(timeout=0.1) == suback_packet
    assert handle2.wait_for_ack(timeout=0.1) is None

    suback_packet.packet_id = 2
    subscriptions.handle_suback(suback_packet)

    assert handle2.wait_for_ack(timeout=0.1) == suback_packet

    subscriptions.handle_connack(MQTTConnAckPacket(session_present=session_present))

    if not session_present:
        # If the session was not present, we should send the SUBSCRIBE packets again.
        assert mock_connection.send.call_count == 2
        assert mock_connection.send.call_args_list[0][0][0] == MQTTSubscribePacket(
            topics=[("test/topic", 2)],
            packet_id=1,
        )
        assert mock_connection.send.call_args_list[1][0][0] == MQTTSubscribePacket(
            topics=[("test/topic", 2)],
            packet_id=2,
        )
        mock_connection.reset_mock()


@pytest.mark.parametrize(("tf", "topic"), [
    ("test/topic", "test/topic"),
    ("test/+", "test/topic"),
    ("#", "test/topic"),
])
def test_subscriptions_handle_publish(
    tf: str,
    topic: str,
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test handling a publish packet."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    recvd = []
    def callback(client: Client, packet: MQTTPublishPacket) -> None:
        recvd.append(packet)
    subscriptions.subscribe(tf, callback)

    publish_packet = MQTTPublishPacket(
        topic=topic,
        payload=b"test payload",
    )
    subscriptions.handle_publish(publish_packet)

    assert recvd == [publish_packet]


def test_subscriptions_slots(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))
    # Slots all the way down.
    assert not hasattr(subscriptions, "__dict__")
    assert hasattr(subscriptions, "__weakref__")
    assert all(hasattr(subscriptions, attr) for attr in subscriptions.__slots__)



from __future__ import annotations

import ssl
import threading
from typing import Final, Iterable, Sequence
import weakref

from .connection import Address, ConnectParams, Connection, MessageHandlers
from .logger import get_logger
from .mqtt_spec import MQTTReasonCode
from .packet import MQTTAuthPacket
from .property import MQTTAuthProps, MQTTConnectProps, MQTTPublishProps, MQTTWillProps
from .persistence.base import PublishHandle
from .session import Session
from .subscriptions import Subscriptions, SubscribeCallback, SubscribeHandle, UnsubscribeHandle, RetainPolicy
from .topic_alias import AliasPolicy

logger: Final = get_logger("client")


class Client:
    """High level interface for the MQTT client.

    :param db_path: Path to the database file for persistence.
    :param db_fast: If True, use a faster database implementation (e.g. SQLite WAL).
    """
    __slots__ = (
        "__weakref__",
        "_thread",
        "connection",
        "session",
    )

    def __init__(self, db_path: str = "", *, db_fast: bool = False) -> None:
        self._thread: threading.Thread | None = None
        message_handlers = MessageHandlers()
        with message_handlers as handlers:
            self.connection = Connection(handlers)
            subscriptions = Subscriptions(handlers, self.connection, weakref.ref(self))
            self.session = Session(handlers, subscriptions, self.connection, db_path=db_path, db_fast=db_fast)
            handlers.register(MQTTAuthPacket, self.handle_auth)

    def __enter__(self) -> Client:
        self.start_loop()
        return self

    def __exit__(self, *args: object) -> None:
        self.shutdown()

    def connect(
        self,
        address: str,
        *,
        client_id: str = "",
        clean_start: bool = False,
        connect_timeout: float | None = None,
        reconnect_delay: int = 0,
        keepalive_interval: int = 0,
        tcp_nodelay: bool = True,
        tls_context: ssl.SSLContext | None = None,
        tls_hostname: str = "",
        will_topic: str = "",
        will_payload: bytes = b"",
        will_qos: int = 0,
        will_retain: bool = False,
        will_properties: MQTTWillProps | None = None,
        connect_properties: MQTTConnectProps | None = None,
    ) -> None:
        """Connect to the broker.

        :param address: The address of the broker to connect to.
        :param client_id: The client ID to use for the connection, or empty string to request one from the broker.
        :param clean_start: If True, an existing session will not be resumed.
        :param connect_timeout: Timeout for the connection attempt in seconds, or None for no timeout.
        :param reconnect_delay: Delay in seconds before attempting to reconnect after a disconnection.
        :param keepalive_interval: The keep alive interval in seconds, or 0 to disable keep alive.
        :param tcp_nodelay: If True, enable TCP_NODELAY to disable Nagle's algorithm.
        :param tls_context: An SSLContext for TLS connections, or None to use the default.
        :param tls_hostname: The hostname to use for TLS connections, or empty string to determine from the address.
        :param will_topic: The topic for the Will message, or empty string to disable.
        :param will_payload: The payload for the Will message.
        :param will_qos: The QoS level for the Will message (0, 1, or 2).
        :param will_retain: If True, the Will message will be retained.
        :param will_properties: Properties for the Will message.
        :param connect_properties: Properties for the CONNECT packet.
        """
        _address = Address(address)
        params = ConnectParams(
            address=_address,
            client_id=client_id,
            clean_start=clean_start,
            connect_timeout=connect_timeout,
            reconnect_delay=reconnect_delay,
            keepalive_interval=keepalive_interval,
            tcp_nodelay=tcp_nodelay,
            tls_context=tls_context,
            tls_hostname=tls_hostname,
            will_topic=will_topic,
            will_payload=will_payload,
            will_qos=will_qos,
            will_retain=will_retain,
            will_properties=will_properties if will_properties is not None else MQTTWillProps(),
            connect_properties=connect_properties if connect_properties is not None else MQTTConnectProps(),
        )
        self.session.set_params(params)
        self.connection.connect(params)

    def disconnect(self) -> None:
        """Disconnect from the broker."""
        self.connection.disconnect()

    def shutdown(self) -> None:
        """Shutdown the client and close the connection."""
        self.connection.shutdown()

    def publish(
        self,
        topic: str,
        payload: bytes,
        *,
        qos: int = 0,
        retain: bool = False,
        properties: MQTTPublishProps | None = None,
        alias_policy: AliasPolicy = AliasPolicy.NEVER,
    ) -> PublishHandle:
        """Publish a message to a topic.

        :param topic: The topic to publish to.
        :param payload: The payload of the message.
        :param qos: The QoS level for the message (0, 1, or 2).
        :param retain: If True, the message will be retained by the broker.
        :param properties: Properties for the PUBLISH packet.
        :param alias_policy: The policy for using automatic topic aliases.
        """
        properties = properties if properties is not None else None
        return self.session.publish(
            topic,
            payload,
            qos=qos,
            retain=retain,
            properties=properties,
            alias_policy=alias_policy,
        )

    def subscribe(
        self,
        topic_filter: str,
        callback: SubscribeCallback,
        max_qos: int = 2,
        *,
        share_name: str | None = None,
        no_local: bool = False,
        retain_as_published: bool = False,
        retain_policy: RetainPolicy = RetainPolicy.ALWAYS,
        sub_id: int | None = None,
        user_properties: Sequence[tuple[str, str]] | None = None,
    ) -> SubscribeHandle | None:
        """Subscribe to a topic filter with a callback.

        If the client is connected, returns a handle which can be used to unsubscribe from the topic filter
        or wait for the subscription to be acknowledged.

        If the client is not connected, returns None.

        :param topic_filter: The topic filter to subscribe to.
        :param callback: The callback to call when a message is received on the subscribed topic.
        :param max_qos: The maximum QoS level for the subscription (0, 1, or 2).
        :param share_name: The name of a shared subscription to use.
        :param no_local: If True, do not receive messages published by this client.
        :param retain_as_published: If True, the retain flag of messages will match the original message.
        :param retain_policy: The policy for retained messages.
        :param sub_id: An optional subscription ID for the subscription.
        :param user_properties: Optional user properties to include in the subscription request.
        """
        return self.session.subscriptions.subscribe(
            topic_filter,
            callback,
            max_qos=max_qos,
            share_name=share_name,
            no_local=no_local,
            retain_as_published=retain_as_published,
            retain_policy=retain_policy,
            sub_id=sub_id,
            user_properties=user_properties,
        )

    def unsubscribe(
        # This method must have the same signature as the subscribe method.
        # This lets us match the unsubscribe to the subscribe with the same args.
        self,
        topic_filter: str,
        callback: SubscribeCallback,
        max_qos: int = 2,
        *,
        share_name: str | None = None,
        no_local: bool = False,
        retain_as_published: bool = False,
        retain_policy: RetainPolicy = RetainPolicy.ALWAYS,
        sub_id: int | None = None,
        user_properties: Iterable[tuple[str, str]] | None = None,
        unsub_user_properties: Iterable[tuple[str, str]] | None = None,
    ) -> UnsubscribeHandle | None:
        """Unsubscribe from a topic filter.

        If the client is connected, returns a handle which can be used to wait for the unsubscription to be acknowledged.

        If the client is not connected, returns None.

        :param topic_filter: The topic filter to unsubscribe from.
        :param callback: The callback that was used to subscribe to the topic filter.
        :param max_qos: The maximum QoS level for the unsubscription (0, 1, or 2).
        :param share_name: The name of a shared subscription to use.
        :param no_local: If True, do not receive messages published by this client.
        :param retain_as_published: If True, the retain flag of messages will match the original message.
        :param retain_policy: The policy for retained messages.
        :param sub_id: An optional subscription ID for the unsubscription.
        :param user_properties: Optional user properties which were included in the subscription request.
        :param unsub_user_properties: Optional user properties to include in the unsubscription request.
        """
        return self.session.subscriptions.unsubscribe(
            topic_filter,
            callback,
            max_qos=max_qos,
            share_name=share_name,
            no_local=no_local,
            retain_as_published=retain_as_published,
            retain_policy=retain_policy,
            sub_id=sub_id,
            user_properties=user_properties,
            unsub_user_properties=unsub_user_properties,
        )

    def auth(
        self,
        *,
        authentication_method: str | None = None,
        authentication_data: bytes | None = None,
        reason_string: str | None = None,
        user_properties: Sequence[tuple[str, str]] | None = None,
        reason_code: MQTTReasonCode = MQTTReasonCode.Success,
    ) -> None:
        """Send an AUTH packet to the broker.

        :param authentication_method: The authentication method to use.
        :param authentication_data: Authentication data to send.
        :param reason_string: A reason string to include in the AUTH packet.
        :param user_properties: Optional user properties to include in the AUTH packet.
        :param reason_code: The reason code for the AUTH packet.
        """
        properties = MQTTAuthProps()
        if authentication_method is not None:
            properties.AuthenticationMethod = authentication_method
        if authentication_data is not None:
            properties.AuthenticationData = authentication_data
        if reason_string is not None:
            properties.ReasonString = reason_string
        if user_properties is not None:
            properties.UserProperty = user_properties
        packet = MQTTAuthPacket(
            reason_code=reason_code,
            properties=properties,
        )
        self.connection.send(packet)

    def wait_for_connect(self, timeout: float | None = None) -> None:
        """Wait for the client to connect to the broker.

        :raises TimeoutError: The timeout was exceeded."""
        if not self.connection.wait_for_connect(timeout):
            raise TimeoutError("Waiting for connection timed out")

    def wait_for_disconnect(self, timeout: float | None = None) -> None:
        """Wait for the client to disconnect from the broker.

        :raises TimeoutError: The timeout was exceeded."""
        if not self.connection.wait_for_disconnect(timeout):
            raise TimeoutError("Waiting for disconnection timed out")

    def wait_for_shutdown(self, timeout: float | None = None) -> None:
        """Wait for the client to disconnect and finalize.

        :raises TimeoutError: The timeout was exceeded."""
        if not self.connection.wait_for_shutdown(timeout):
            raise TimeoutError("Waiting for disconnection timed out")

    def start_loop(self) -> None:
        """Start the client state machine in a separate thread."""
        if self._thread is not None:
            raise RuntimeError("Connection loop already started")
        self._thread = threading.Thread(target=self.loop_forever, daemon=True)
        self._thread.start()

    def loop_once(self, max_wait: float | None = 0.0) -> None:
        """Run a single iteration of the MQTT client loop.

        If max_wait is 0.0 (the default), this call will not block.

        If max_wait is None, this call will block until the next event.

        Any other numeric max_wait value may block for maximum that amount of time in seconds."""
        self.connection.loop_once(max_wait)

    def loop_forever(self) -> None:
        """Run the MQTT client loop.

        This will run until the client is stopped or shutdown.
        """
        self.connection.loop_forever()

    def loop_until_connected(self, timeout: float | None = None) -> bool:
        """Run the MQTT client loop until the client is connected to the broker.

        If a timeout is provided, the loop will give up after that amount of time.

        Returns True if the client is connected, False if the timeout was reached."""
        return self.connection.loop_until_connected(timeout)

    def is_connected(self) -> bool:
        """Check if the client is connected to the broker."""
        return self.connection.is_connected()

    def handle_auth(self, packet: MQTTAuthPacket) -> None:
        """Callback for an AUTH packet from the broker."""
        logger.debug("Got an AUTH packet")



from contextlib import suppress
import sqlite3
from threading import Condition
from typing import Final
import weakref

from .base import Persistence, ReliablePublishHandle, RenderedPacket
from ..logger import get_logger
from ..mqtt_spec import MAX_PACKET_ID
from ..packet import MQTTPublishPacket, MQTTPubRelPacket
from ..property import MQTTPublishProps
from ..topic_alias import AliasPolicy

logger: Final = get_logger("persistence.sqlite")


class SQLitePersistence(Persistence):
    """SQLite persistence for MQTT messages.

    This class provides a SQLite-based persistence layer for MQTT messages.
    """
    __slots__ = ("_cond", "_conn", "_cursor", "_db_path", "_handles")
    def __init__(self, db_path: str, *, db_fast: bool = False) -> None:
        self._cond = Condition()
        self._handles: weakref.WeakValueDictionary[int, ReliablePublishHandle] = weakref.WeakValueDictionary({})
        self._db_path = db_path
        self._conn = sqlite3.connect(self._db_path, check_same_thread=False)
        self._cursor = self._conn.cursor()
        if db_fast:
            self._cursor.executescript(
                """
                PRAGMA journal_mode = WAL;
                PRAGMA synchronous = NORMAL;
                """
            )
        self._create_tables()

    def __del__(self) -> None:
        # Ignore "already closed" errors while closing resources.
        with suppress(sqlite3.ProgrammingError):
            self._cursor.close()
        with suppress(sqlite3.ProgrammingError):
            self._conn.close()

    def __len__(self) -> int:
        with self._cond:
            self._cursor.execute(
                """
                SELECT COUNT(*) FROM messages
                """
            )
            row = self._cursor.fetchone()
            if row is None:
                return 0
            return int(row[0])

    def _create_tables(self) -> None:
        """Create the necessary tables in the SQLite database."""
        with self._cond:
            self._cursor.executescript(
                """
                BEGIN;
                CREATE TABLE IF NOT EXISTS messages (
                    id INTEGER PRIMARY KEY,
                    topic TEXT NOT NULL,
                    payload BLOB NOT NULL,
                    qos INTEGER NOT NULL,
                    retain INTEGER NOT NULL,
                    properties BLOB,
                    alias_policy INTEGER NOT NULL,
                    dup INTEGER DEFAULT 0,
                    received INTEGER DEFAULT 0,
                    packet_id INTEGER UNIQUE DEFAULT NULL,
                    inflight INTEGER DEFAULT 0
                ) STRICT;
                CREATE TABLE IF NOT EXISTS received (
                    packet_id INTEGER PRIMARY KEY
                ) STRICT;
                CREATE TABLE IF NOT EXISTS client_id (
                    id INTEGER PRIMARY KEY CHECK (id = 0),
                    client_id TEXT DEFAULT NULL
                ) STRICT;
                INSERT OR IGNORE INTO client_id (id) VALUES (0);
                COMMIT;
                """
            )

    def _get_client_id(self) -> str:
        """Get the client ID from the database."""
        with self._cond:
            self._cursor.execute(
                """
                SELECT client_id FROM client_id WHERE id = 0
                """
            )
            row = self._cursor.fetchone()
            if row is None:
                return ""
            return str(row[0])

    def _set_client_id(self, client_id: str) -> None:
        """Set the client ID in the database."""
        with self._cond:
            self._cursor.execute(
                """
                UPDATE client_id SET client_id = ? WHERE id = 0
                """,
                (client_id,),
            )
            self._conn.commit()

    def add(
        self,
        topic: str,
        payload: bytes,
        qos: int,
        retain: bool,
        properties: MQTTPublishProps,
        alias_policy: AliasPolicy,
    ) -> ReliablePublishHandle:
        assert alias_policy != AliasPolicy.ALWAYS, "AliasPolicy must not be ALWAYS for retained messages."
        with self._cond:
            properties_blob = properties.encode() if len(properties) > 0 else None
            self._cursor.execute(
                """
                INSERT INTO messages (topic, payload, qos, retain, properties, alias_policy)
                VALUES (?, ?, ?, ?, ?, ?)
                """,
                (topic, payload, qos, int(retain), properties_blob, alias_policy.value),
            )
            self._conn.commit()
            message_id = self._cursor.lastrowid
            if message_id is None:
                raise ValueError("Failed to retrieve the last inserted row ID.")
            handle = ReliablePublishHandle(self._cond)
            self._handles[message_id] = handle
            return handle

    def get(self, count: int) -> list[int]:
        with self._cond:
            self._cursor.execute(
                """
                SELECT id FROM messages
                WHERE inflight = 0
                ORDER BY id ASC
                LIMIT ?
                """,
                (count,),
            )
            rows = self._cursor.fetchall()
            return [row[0] for row in rows]

    def ack(self, packet_id: int) -> None:
        with self._cond:
            self._cursor.execute(
                """
                SELECT id, qos, received FROM messages WHERE packet_id = ?
                """,
                (packet_id,),
            )
            row = self._cursor.fetchone()
            if row is None:
                raise ValueError(f"Unknown packet_id: {packet_id}")
            message_id, qos, received = row
            if received or qos == 1:
                # If the message is QoS 1, we need to delete it from the store.
                self._cursor.execute(
                    """
                    DELETE FROM messages WHERE id = ?
                    """,
                    (message_id,),
                )
                handle = self._handles.pop(message_id, None)
                if handle is not None:
                    handle.acked = True
                    self._cond.notify_all()
            else:
                # If the message is QoS 2, we need to mark it as received.
                self._cursor.execute(
                    """
                    UPDATE messages SET inflight = 0, received = 1 WHERE id = ?
                    """,
                    (message_id,),
                )
            self._conn.commit()

    def check_rec(self, packet: MQTTPublishPacket) -> bool:
        if packet.qos != 2:
            raise ValueError("Not a QoS 2 PUBLISH packet")
        with self._cond:
            self._cursor.execute(
                """
                SELECT packet_id FROM received WHERE packet_id = ?
                """,
                (packet.packet_id,),
            )
            row = self._cursor.fetchone()
            if row is not None:
                logger.debug("Received duplicate QoS 2 packet with ID %d", packet.packet_id)
                return False
            return True

    def set_rec(self, packet: MQTTPublishPacket) -> None:
        if packet.qos != 2:
            raise ValueError("Not a QoS 2 PUBLISH packet")
        with self._cond:
            self._cursor.execute(
                """
                INSERT INTO received (packet_id) VALUES (?)
                """,
                (packet.packet_id,),
            )
            self._conn.commit()

    def rel(self, packet: MQTTPubRelPacket) -> None:
        with self._cond:
            self._cursor.execute(
                """
                DELETE FROM received WHERE packet_id = ?
                """,
                (packet.packet_id,),
            )
            self._conn.commit()

    def _generate_packet_id(self) -> int:
        """Generate a unique packet ID for the message."""
        with self._cond:
            self._cursor.execute(
                """
                SELECT packet_id from messages WHERE packet_id = (SELECT MAX(packet_id) FROM messages)
                """
            )
            row = self._cursor.fetchone()
            packet_id = 1
            if row is not None:
                packet_id = row[0]
                packet_id = packet_id + 1 if packet_id < MAX_PACKET_ID else 1
            return packet_id

    def render(self, message_id: int) -> RenderedPacket:
        with self._cond:
            self._cursor.execute(
                """
                SELECT topic, payload, qos, retain, properties, dup, received, packet_id, alias_policy
                FROM messages
                WHERE id = ?
                """,
                (message_id,),
            )
            row = self._cursor.fetchone()
            if row is None:
                raise ValueError(f"Message ID {message_id} not found in persistence store.")
            topic, payload, qos, retain, properties_blob, dup, received, packet_id, alias_policy = row
            if properties_blob is not None:
                properties_view = memoryview(properties_blob)
                properties, _ = MQTTPublishProps.decode(properties_view)
            else:
                properties = MQTTPublishProps()
            if packet_id is None:
                packet_id = self._generate_packet_id()
            packet: MQTTPublishPacket | MQTTPubRelPacket
            if received:
                alias_policy = AliasPolicy.NEVER
                packet = MQTTPubRelPacket(packet_id=packet_id)
            else:
                alias_policy = AliasPolicy(alias_policy)
                packet = MQTTPublishPacket(
                    topic=topic,
                    payload=payload,
                    packet_id=packet_id,
                    qos=qos,
                    retain=bool(retain),
                    properties=properties,
                    dup=dup,
                )
            self._cursor.execute(
                """
                UPDATE messages SET inflight = 1, packet_id = ? WHERE id = ?
                """,
                (packet_id, message_id),
            )
            self._conn.commit()
            return RenderedPacket(packet, alias_policy)

    def _reset_inflight(self) -> None:
        """Clear inflight status of all messages."""
        with self._cond:
            self._cursor.executescript(
                """
                BEGIN;
                UPDATE messages SET dup = 1 WHERE inflight = 1;
                UPDATE messages SET inflight = 0;
                COMMIT;
                """
            )

    def clear(self) -> None:
        with self._cond:
            self._cursor.executescript(
                """
                BEGIN;
                DELETE FROM messages;
                DELETE FROM received;
                COMMIT;
                """
            )
            self._handles.clear()

    def open(self, client_id: str, clear: bool = False) -> None:
        logger.debug("Opening SQLite persistence with client ID: %s clear=%s", client_id, clear)
        with self._cond:
            if clear or client_id != self._get_client_id():
                logger.debug("Clearing SQLite persistence for client ID: %s", client_id)
                self._set_client_id(client_id)
                self.clear()
            else:
                self._reset_inflight()



from abc import ABCMeta, abstractmethod
from threading import Condition
from typing import ClassVar, NamedTuple, Sequence

from ..packet import MQTTPublishPacket, MQTTPubRelPacket
from ..property import MQTTPublishProps
from ..topic_alias import AliasPolicy


class PublishHandle(metaclass=ABCMeta):
    """Represents a publish operation."""
    __slots__: ClassVar[Sequence[str]] = tuple()

    @abstractmethod
    def is_acked(self) -> bool:
        """Check if the message has been acknowledged.

        For qos=0, this is always False.
        For qos=1, this is True if the message has been acknowledged.
        For qos=2, this is True if the message has been completely acknowledged."""
        ...  # pragma: no cover

    @abstractmethod
    def wait_for_ack(self, timeout: float | None = None) -> bool:
        """Wait for the message to be acknowledged.

        For qos=0, this always returns False immediately.
        For qos=1, this returns True if the message has been acknowledged.
        For qos=2, this returns True if the message has been completely acknowledged.
        If the timeout is exceeded, this returns False."""
        ...  # pragma: no cover


class UnreliablePublishHandle(PublishHandle):
    """Represents a publish operation with qos=0."""
    __slots__ = tuple()

    def is_acked(self) -> bool:
        return False

    def wait_for_ack(self, timeout: float | None = None) -> bool:
        return False


class ReliablePublishHandle(PublishHandle):
    """Represents a publish operation with qos>0."""
    __slots__ = ("__weakref__", "_cond", "acked")

    def __init__(self, cond: Condition) -> None:
        self._cond = cond
        self.acked = False

    def is_acked(self) -> bool:
        return self.acked

    def wait_for_ack(self, timeout: float | None = None) -> bool:
        with self._cond:
            self._cond.wait_for(self.is_acked, timeout)
        return self.acked


class RenderedPacket(NamedTuple):
    """Represents a rendered packet."""
    packet: MQTTPublishPacket | MQTTPubRelPacket
    alias_policy: AliasPolicy


class Persistence(metaclass=ABCMeta):
    """Abstract base class for message persistence."""
    __slots__: ClassVar[Sequence[str]] = tuple()

    @abstractmethod
    def __len__(self) -> int:
        """Return the number of outgoing messages in the persistence store."""
        ...  # pragma: no cover

    @abstractmethod
    def add(
        self,
        topic: str,
        payload: bytes,
        qos: int,
        retain: bool,
        properties: MQTTPublishProps,
        alias_policy: AliasPolicy,
    ) -> ReliablePublishHandle:
        """Add a PUBLISH message to the persistence store."""
        ...  # pragma: no cover

    @abstractmethod
    def get(self, count: int) -> Sequence[int]:
        """Get the packet ids of some pending messages from the store."""
        ...  # pragma: no cover

    @abstractmethod
    def ack(self, packet_id: int) -> None:
        """Ack a PUBLISH or PUBREL message in the persistence store.

        Raises ValueError if the packet_id is not found in the store."""
        ...  # pragma: no cover

    @abstractmethod
    def check_rec(self, packet: MQTTPublishPacket) -> bool:
        """Validate that a QoS 2 PUBLISH packet has not already been received.

        Returns True if the packet has not already been received, otherwise False.

        Raises ValueError if the packet is not a QoS 2 PUBLISH packet."""
        ...  # pragma: no cover

    @abstractmethod
    def set_rec(self, packet: MQTTPublishPacket) -> None:
        """Indicate that a QoS 2 PUBLISH message has been received.

        Raises ValueError if the packet is not a QoS 2 PUBLISH packet."""
        ...  # pragma: no cover

    @abstractmethod
    def rel(self, packet: MQTTPubRelPacket) -> None:
        """Release a QoS 2 PUBLISH message."""
        ...  # pragma: no cover

    @abstractmethod
    def render(self, packet_id: int) -> RenderedPacket:
        """Render a PUBLISH message from the persistence store.

        This also indicates to the persistence store that the message is inflight."""
        ...  # pragma: no cover

    @abstractmethod
    def clear(self) -> None:
        """Clear the persistence store, destroying all pending messages."""
        ...  # pragma: no cover

    @abstractmethod
    def open(self, client_id: str, clear: bool = False) -> None:
        """Indicate to the persistence store that the broker has acknowledged our connection.

        This may clear the persistence store if the client_id is different from the persisted,
        or if clear is True."""
        ...  # pragma: no cover



from __future__ import annotations

from collections import deque
from enum import IntEnum
from threading import Condition
from typing import Callable, Final, Iterable, NamedTuple, Sequence, TYPE_CHECKING
import weakref

from .connection import Connection, InvalidStateError, MessageHandlers
from .logger import get_logger
from .mqtt_spec import MAX_PACKET_ID
from .packet import (
    MQTTPublishPacket,
    MQTTSubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubscribePacket,
    MQTTUnsubAckPacket,
    MQTTConnAckPacket,
)
from .property import MQTTSubscribeProps, MQTTUnsubscribeProps
from .topic_alias import InboundTopicAlias
from .topic_filter import validate_topic_filter, validate_share_name, join_share, match_topic_filter

if TYPE_CHECKING:
    from .client import Client  # pragma: no cover

logger: Final = get_logger("subscriptions")

SubscribeCallback = Callable[["Client", MQTTPublishPacket], None]


class NoMatchingSubscriptionError(Exception):
    """Exception raised when no matching subscription is found when unsubscribing."""


class RetainPolicy(IntEnum):
    """Policy for broker sending retained messages upon a subscription.

    ALWAYS: Always send retained messages on subscription.

    ONCE: Only send retained messages on the first subscription to a topic.

    NEVER: Never send retained messages on subscription."""
    ALWAYS = 0
    ONCE = 1
    NEVER = 2


class SubscribeHandle:
    """Represents a subscription to a topic filter with a callback."""
    __slots__ = ("_data", "_subscriptions", "ack", "failed")

    ack: MQTTSubAckPacket | None
    failed: bool
    _data: Subscription
    _subscriptions: weakref.ReferenceType[Subscriptions]

    def __init__(self, data: Subscription, subscriptions: weakref.ReferenceType[Subscriptions]) -> None:
        self._data = data
        self._subscriptions = subscriptions
        self.ack = None
        self.failed = False

    def unsubscribe(self) -> UnsubscribeHandle | None:
        """Unsubscribe from the topic filter.

        Returns an UnsubscribeHandle if the operation was successful, otherwise None."""
        subscriptions = self._subscriptions()
        if subscriptions is not None:
            return subscriptions.unsubscribe(**(self._data._asdict()))
        return None

    def wait_for_ack(self, timeout: float | None = None) -> MQTTSubAckPacket | None:
        """Wait for the subscription acknowledgement.

        Returns the acknowledgement packet if the subscription was sent to the broker and acknowledged.

        Returns None if the subscription was not sent to the broker or timeout was reached."""
        subscriptions = self._subscriptions()
        if subscriptions is not None:
            return subscriptions.wait_for_suback(self, timeout=timeout)
        return None


class UnsubscribeHandle:
    """Represents an unsubscribe request which was sent to the broker."""
    __slots__ = ("_subscriptions", "ack", "failed")

    ack: MQTTUnsubAckPacket | None
    failed: bool
    _subscriptions: weakref.ReferenceType[Subscriptions]

    def __init__(self, subscriptions: weakref.ReferenceType[Subscriptions]) -> None:
        self.ack = None
        self.failed = False
        self._subscriptions = subscriptions

    def wait_for_ack(self, timeout: float | None = None) -> MQTTUnsubAckPacket | None:
        """Wait for the unsubscribe acknowledgement.

        Returns None if the unsubscribe was not sent to the broker or timeout was reached."""
        subscriptions = self._subscriptions()
        if subscriptions is not None:
            return subscriptions.wait_for_unsuback(self, timeout=timeout)
        return None


class Subscription(NamedTuple):
    """All the data about a request for subscription."""
    topic_filter: str
    max_qos: int
    share_name: str | None
    no_local: bool
    retain_as_published: bool
    retain_policy: RetainPolicy
    sub_id: int | None
    user_properties: tuple[tuple[str, str], ...] | None
    callback: SubscribeCallback

    def render(self) -> MQTTSubscribePacket:
        """Render the subscription into a SUBSCRIBE packet."""
        opts = self.max_qos | (self.retain_policy << 4) | (self.retain_as_published << 3) | (self.no_local << 2)
        props = MQTTSubscribeProps()
        if self.sub_id is not None:
            props.SubscriptionIdentifier = {self.sub_id}
        if self.user_properties is not None:
            props.UserProperty = self.user_properties
        topic_filter = join_share(self.topic_filter, self.share_name)
        return MQTTSubscribePacket(
            topics=[(topic_filter, opts)],
            properties=props,
        )


class Subscriptions:
    """Container for MQTT subscriptions and their callbacks."""
    __slots__ = (
        "__weakref__",
        "_client",
        "_cond",
        "_connection",
        "_next_sub_packet_id",
        "_next_unsub_packet_id",
        "_out_of_session",
        "_sub_handles",
        "_subscriptions",
        "_topic_alias",
        "_unsub_handles",
    )

    def __init__(
        self,
        handlers: MessageHandlers,
        connection: Connection,
        client: weakref.ReferenceType[Client],
    ) -> None:
        self._connection = connection
        self._client = client
        self._cond = Condition()
        self._topic_alias = InboundTopicAlias()
        self._subscriptions: dict[str, list[Subscription]] = {}
        self._sub_handles: dict[int, SubscribeHandle] = {}
        self._unsub_handles: dict[int, UnsubscribeHandle] = {}
        self._next_sub_packet_id = 1
        self._next_unsub_packet_id = 1
        self._out_of_session: deque[MQTTSubscribePacket | MQTTUnsubscribePacket] = deque()

        handlers.register(MQTTSubAckPacket, self.handle_suback)
        handlers.register(MQTTUnsubAckPacket, self.handle_unsuback)
        handlers.register(MQTTConnAckPacket, self.handle_connack)

    def subscribe(
        self,
        topic_filter: str,
        callback: SubscribeCallback,
        max_qos: int = 2,
        *,
        share_name: str | None = None,
        no_local: bool = False,
        retain_as_published: bool = False,
        retain_policy: RetainPolicy = RetainPolicy.ALWAYS,
        sub_id: int | None = None,
        user_properties: Sequence[tuple[str, str]] | None = None,
    ) -> SubscribeHandle | None:
        """Add a subscription with a callback.

        If successful, returns a SubscribeHandle which can be used to unsubscribe the callback.

        Returns None if the subscription was not sent to the broker."""
        validate_topic_filter(topic_filter)
        if share_name is not None:
            validate_share_name(share_name)
        sub = Subscription(
            topic_filter=topic_filter,
            max_qos=max_qos,
            share_name=share_name,
            no_local=no_local,
            retain_as_published=retain_as_published,
            retain_policy=retain_policy,
            sub_id=sub_id,
            user_properties=tuple(user_properties) if user_properties is not None else None,
            callback=callback,
        )
        with self._cond:
            if topic_filter not in self._subscriptions:
                self._subscriptions[topic_filter] = []
            self._subscriptions[topic_filter].append(sub)
            packet = sub.render()
            packet.packet_id = self._get_next_sub_packet_id()
            try:
                self._connection.send(packet)
            except InvalidStateError:
                logger.debug("Connection not ready, SUBSCRIBE not sent")
                self._out_of_session.append(packet)
                return None
            else:
                handle = SubscribeHandle(sub, weakref.ref(self))
                self._sub_handles[packet.packet_id] = handle
                return handle

    def _get_next_sub_packet_id(self) -> int:
        """Get the next SUBSCRIBE packet identifier."""
        sub_id = self._next_sub_packet_id
        self._next_sub_packet_id += 1
        if self._next_sub_packet_id > MAX_PACKET_ID:
            self._next_sub_packet_id = 1
        return sub_id

    def wait_for_suback(
        self,
        handle: SubscribeHandle,
        timeout: float | None = None,
    ) -> MQTTSubAckPacket | None:
        """Wait for a SUBACK for a SubscribeHandle.

        Returns the SUBACK packet if the subscription was sent to the broker and acknowledged.

        Returns None if the subscription was not sent to the broker or timeout was reached."""
        with self._cond:
            if self._cond.wait_for(lambda: handle.ack is not None or handle.failed, timeout=timeout):
                return handle.ack
            return None

    def unsubscribe(
        # This method must have the same signature as the subscribe method.
        # This lets us match the unsubscribe to the subscribe with the same args.
        self,
        topic_filter: str,
        callback: SubscribeCallback,
        max_qos: int = 2,
        *,
        share_name: str | None = None,
        no_local: bool = False,
        retain_as_published: bool = False,
        retain_policy: RetainPolicy = RetainPolicy.ALWAYS,
        sub_id: int | None = None,
        user_properties: Iterable[tuple[str, str]] | None = None,
        unsub_user_properties: Iterable[tuple[str, str]] | None = None,
    ) -> UnsubscribeHandle | None:
        """Unsubscribe from a topic filter.

        Returns an UnsubscribeHandle if the operation was successful, otherwise None."""
        with self._cond:
            if topic_filter not in self._subscriptions:
                # Nothing to unsubscribe.
                return None
            user_props = tuple(user_properties) if user_properties is not None else None
            matches = [
                sub for sub in self._subscriptions[topic_filter]
                if (
                    sub.share_name == share_name
                    and sub.callback == callback
                    and sub.max_qos == max_qos
                    and sub.no_local is no_local
                    and sub.retain_as_published is retain_as_published
                    and sub.retain_policy == retain_policy
                    and sub.sub_id == sub_id
                    and sub.user_properties == user_props
                )
            ]
            for match in matches:
                self._subscriptions[topic_filter].remove(match)
            remaining = len([sub for sub in self._subscriptions[topic_filter] if sub.share_name == share_name])
            if remaining > 0:
                # There are still subscriptions left for this topic filter.
                return None
            # Nothing left, send UNSUBSCRIBE.
            del self._subscriptions[topic_filter]
            packet = MQTTUnsubscribePacket(
                topics=[join_share(topic_filter, share_name)],
                packet_id=self._get_next_unsub_packet_id(),
                properties=MQTTUnsubscribeProps(),
            )
            if unsub_user_properties is not None:
                packet.properties.UserProperty = tuple(unsub_user_properties)
            try:
                self._connection.send(packet)
            except InvalidStateError:
                logger.debug("Connection closed, UNSUBSCRIBE not sent")
                self._out_of_session.append(packet)
                return None
            else:
                handle = UnsubscribeHandle(weakref.ref(self))
                self._unsub_handles[packet.packet_id] = handle
                return handle

    def _get_next_unsub_packet_id(self) -> int:
        """Get the next UNSUBSCRIBE packet identifier."""
        unsub_id = self._next_unsub_packet_id
        self._next_unsub_packet_id += 1
        if self._next_unsub_packet_id > MAX_PACKET_ID:
            self._next_unsub_packet_id = 1
        return unsub_id

    def wait_for_unsuback(
        self,
        handle: UnsubscribeHandle,
        timeout: float | None = None,
    ) -> MQTTUnsubAckPacket | None:
        """Wait for a UNSUBACK for an UnsubscribeHandle.

        Returns the UNSUBACK packet if the subscription was sent to the broker and acknowledged.

        Returns None if the subscription was not sent to the broker or timeout was reached."""
        with self._cond:
            if self._cond.wait_for(lambda: handle.ack is not None or handle.failed, timeout=timeout):
                return handle.ack
            return None

    def handle_publish(self, packet: MQTTPublishPacket) -> None:
        """Handle incoming PUBLISH packets."""
        with self._cond:
            client = self._client()
            if client is None:
                raise RuntimeError("Client went out of scope")
            self._topic_alias.handle(packet)
            sub_lists = [sub for tf, sub in self._subscriptions.items() if match_topic_filter(tf, packet.topic)]
            subs = [sub for sub_list in sub_lists for sub in sub_list]
            if packet.properties.SubscriptionIdentifier is not None:
                subs = [sub for sub in subs if sub.sub_id in packet.properties.SubscriptionIdentifier]
            for sub in subs:
                try:
                    sub.callback(client, packet)
                except Exception:  # noqa: PERF203
                    logger.exception("Unhandled exception in subscription callback")

    def handle_suback(self, packet: MQTTSubAckPacket) -> None:
        """Handle incoming SUBACK packets."""
        if any(True for code in packet.reason_codes if code.is_error()):
            errs = [hex(code) for code in packet.reason_codes if code.is_error()]
            logger.error("Errors found in SUBACK return: %s", errs)
        with self._cond:
            handle = self._sub_handles.pop(packet.packet_id, None)
            if handle is not None:
                handle.ack = packet
                self._cond.notify_all()

    def handle_unsuback(self, packet: MQTTUnsubAckPacket) -> None:
        """Handle incoming UNSUBACK packets."""
        if any(True for code in packet.reason_codes if code.is_error()):
            errs = [hex(code) for code in packet.reason_codes if code.is_error()]
            logger.error("Errors found in UNSUBACK return: %s", errs)
        with self._cond:
            handle = self._unsub_handles.pop(packet.packet_id, None)
            if handle is not None:
                handle.ack = packet
                self._cond.notify_all()

    def handle_connack(self, packet: MQTTConnAckPacket) -> None:
        """Handle incoming CONNACK packets."""
        with self._cond:
            self._reset_connection_states()
            if not packet.session_present:
                # Replay all known subscriptions.
                self._replay_session_not_present()
            else:
                # Replay only packets which were not sent.
                self._replay_session_present()

    def _reset_connection_states(self) -> None:
        """Reset all connection-level states."""
        with self._cond:
            self._next_sub_packet_id = 1
            self._next_unsub_packet_id = 1
            self._topic_alias.reset()
            for sub_handle in self._sub_handles.values():
                if sub_handle.ack is None:
                    sub_handle.failed = True
            self._sub_handles.clear()
            for unsub_handle in self._unsub_handles.values():
                if unsub_handle.ack is None:
                    unsub_handle.failed = True
            self._unsub_handles.clear()
            self._cond.notify_all()  # Should clear all waiting for handles.

    def _replay_session_not_present(self) -> None:
        """Replay all subscriptions."""
        with self._cond:
            self._out_of_session.clear()
            for topic_filter in self._subscriptions:
                for sub in self._subscriptions[topic_filter]:
                    sub_packet = sub.render()
                    sub_packet.packet_id = self._get_next_sub_packet_id()
                    self._connection.send(sub_packet)

    def _replay_session_present(self) -> None:
        """Replay all subscriptions which were not sent to the broker."""
        with self._cond:
            while self._out_of_session:
                packet = self._out_of_session.popleft()
                if isinstance(packet, MQTTSubscribePacket):
                    packet.packet_id = self._get_next_sub_packet_id()
                else:
                    packet.packet_id = self._get_next_unsub_packet_id()
                try:
                    self._connection.send(packet)
                except InvalidStateError:
                    logger.debug("Connection not ready, packet not sent")
                    self._out_of_session.appendleft(packet)
                    break



from __future__ import annotations

import socket
import ssl
from threading import Condition
from typing import Any
from unittest.mock import Mock

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection import Address
from ohmqtt.connection.decoder import IncrementalDecoder, ClosedSocketError
from ohmqtt.connection.fsm import FSM
from ohmqtt.connection.keepalive import KeepAlive
from ohmqtt.connection.selector import InterruptibleSelector
from ohmqtt.connection.states import (
    ConnectingState,
    TLSHandshakeState,
    MQTTHandshakeConnectState,
    MQTTHandshakeConnAckState,
    ConnectedState,
    ReconnectWaitState,
    ClosingState,
    ClosedState,
    ShutdownState,
)
from ohmqtt.connection.timeout import Timeout
from ohmqtt.connection.types import ConnectParams, StateData, StateEnvironment
from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTReasonCode
from ohmqtt.packet import (
    decode_packet,
    MQTTConnectPacket,
    MQTTConnAckPacket,
    MQTTDisconnectPacket,
    MQTTPublishPacket,
    MQTTPingReqPacket,
    MQTTPingRespPacket,
    PING,
    PONG,
)
from ohmqtt.platform import HAS_AF_UNIX
from ohmqtt.property import MQTTConnectProps, MQTTConnAckProps, MQTTWillProps


class EnvironmentCallbacks:
    """Container for StateEnvironment callbacks."""
    def __init__(self, mocker: MockerFixture) -> None:
        self.packet = mocker.Mock()

    def reset(self) -> None:
        self.packet.reset_mock()

    def assert_not_called(self) -> None:
        self.packet.assert_not_called()


@pytest.fixture
def callbacks(mocker: MockerFixture) -> EnvironmentCallbacks:
    return EnvironmentCallbacks(mocker)

@pytest.fixture
def env(callbacks: EnvironmentCallbacks) -> StateEnvironment:
    return StateEnvironment(
        packet_callback=callbacks.packet,
    )

@pytest.fixture
def mock_socket(mocker: MockerFixture) -> Mock:
    mock_socket = mocker.Mock(spec=ssl.SSLSocket)
    mocker.patch("ohmqtt.connection.states._get_socket", return_value=mock_socket)
    return mock_socket  # type: ignore[no-any-return]

@pytest.fixture
def decoder(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=IncrementalDecoder)  # type: ignore[no-any-return]

@pytest.fixture
def mock_keepalive(mocker: MockerFixture) -> Mock:
    mock_keepalive = mocker.Mock(spec=KeepAlive)
    mock_keepalive.get_next_timeout.return_value = 1
    mock_keepalive.should_close.return_value = False
    mock_keepalive.should_send_ping.return_value = False
    return mock_keepalive  # type: ignore[no-any-return]

@pytest.fixture
def mock_timeout(mocker: MockerFixture) -> Mock:
    mock_timeout = mocker.Mock(spec=Timeout)
    mock_timeout.interval = None
    mock_timeout.get_timeout.return_value = 1
    mock_timeout.exceeded.return_value = False
    return mock_timeout  # type: ignore[no-any-return]

@pytest.fixture
def state_data(decoder: Mock, mock_socket: Mock, mock_keepalive: Mock, mock_timeout: Mock) -> StateData:
    data = StateData()
    data.decoder = decoder
    data.keepalive = mock_keepalive
    data.sock = mock_socket
    data.timeout = mock_timeout
    data.connack = MQTTConnAckPacket()
    return data

@pytest.fixture
def params() -> ConnectParams:
    return ConnectParams(address=Address("mqtt://testhost"))

@pytest.fixture
def mock_select(mocker: MockerFixture) -> Mock:
    mock_select = mocker.MagicMock(spec=InterruptibleSelector)
    mock_select.__enter__.return_value = mock_select
    mock_select.select.return_value = (False, False)
    mocker.patch("ohmqtt.connection.fsm.InterruptibleSelector", return_value=mock_select)
    return mock_select  # type: ignore[no-any-return]

@pytest.fixture
def mock_read(mocker: MockerFixture) -> Mock:
    mock_read = mocker.patch.object(ConnectedState, "read_packet", autospec=True)
    mock_read.return_value = False
    return mock_read


@pytest.mark.parametrize("address", ["mqtt://testhost", "mqtts://testhost", "unix:///testpath"])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connecting_happy_path(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    decoder: Mock,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock
) -> None:
    if address.startswith("unix:") and not HAS_AF_UNIX:
        pytest.skip("Unix socket not supported on this platform")
    params = ConnectParams(address=Address(address))
    fsm = FSM(env=env, init_state=ConnectingState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # Enter state.
    ConnectingState.enter(fsm, state_data, env, params)
    if params.address.scheme != "unix":
        mock_socket.setsockopt.assert_called_once_with(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    mock_socket.reset_mock()
    decoder.reset.assert_called_once()
    decoder.reset_mock()
    assert mock_timeout.interval == params.connect_timeout
    mock_timeout.mark.assert_called_once()
    mock_timeout.reset_mock()
    assert state_data.disconnect_rc is None
    mock_select.change_sock.assert_called_once_with(mock_socket)
    assert fsm.state is ConnectingState

    # First handle, begin non-blocking connect.
    mock_select.select.return_value = (False, False)
    mock_socket.connect.side_effect = BlockingIOError
    ret = ConnectingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    if params.address.scheme == "unix":
        mock_socket.connect.assert_called_once_with(params.address.host)
    else:
        mock_socket.connect.assert_called_once_with((params.address.host, params.address.port))
    mock_socket.reset_mock()
    assert fsm.state is ConnectingState

    # Second handle, finish non-blocking connect.
    mock_select.select.return_value = (False, True)
    ret = ConnectingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_select.select.assert_called_once_with(write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_socket.setblocking.assert_called_once_with(False)
    mock_socket.reset_mock()
    if params.address.use_tls:
        assert fsm.state is TLSHandshakeState
    else:
        assert fsm.state is MQTTHandshakeConnectState

    callbacks.assert_not_called()


@pytest.mark.parametrize("address", ["mqtt://testhost", "mqtts://testhost", "unix:///testpath"])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connecting_timeout(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_timeout: Mock
) -> None:
    if address.startswith("unix:") and not HAS_AF_UNIX:
        pytest.skip("Unix socket not supported on this platform")
    params = ConnectParams(address=Address(address))
    fsm = FSM(env=env, init_state=ConnectingState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    mock_timeout.exceeded.return_value = True
    ConnectingState.enter(fsm, state_data, env, params)
    mock_timeout.mark.assert_called_once()
    ret = ConnectingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("address", ["mqtt://testhost", "mqtts://testhost", "unix:///testpath"])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connecting_error(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    if address.startswith("unix:") and not HAS_AF_UNIX:
        pytest.skip("Unix socket not supported on this platform")
    params = ConnectParams(address=Address(address))
    fsm = FSM(env=env, init_state=ConnectingState, error_state=ShutdownState)

    mock_socket.connect.side_effect = ConnectionError("TEST")
    ConnectingState.enter(fsm, state_data, env, params)
    ret = ConnectingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_tls_handshake_happy_path(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock,
    mocker: MockerFixture
) -> None:
    params = ConnectParams(address=Address("mqtts://testhost"), tls_context=mocker.Mock())
    fsm = FSM(env=env, init_state=TLSHandshakeState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # Enter state.
    params.tls_context.wrap_socket.return_value = mock_socket  # type: ignore[union-attr]
    TLSHandshakeState.enter(fsm, state_data, env, params)
    params.tls_context.wrap_socket.assert_called_once_with(  # type: ignore[union-attr]
        mock_socket,
        server_hostname=params.address.host,
        do_handshake_on_connect=False,
    )
    mock_select.change_sock.assert_called_once_with(mock_socket)

    # First handle, want write.
    mock_socket.do_handshake.side_effect = ssl.SSLWantWriteError
    mock_select.select.return_value = (False, True)
    ret = TLSHandshakeState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_socket.do_handshake.assert_called_once()
    mock_socket.reset_mock()
    assert fsm.state is TLSHandshakeState

    # Second handle, want read.
    mock_socket.do_handshake.side_effect = ssl.SSLWantReadError
    mock_select.select.return_value = (True, False)
    ret = TLSHandshakeState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(read=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_socket.do_handshake.assert_called_once()
    mock_socket.reset_mock()
    assert fsm.state is TLSHandshakeState

    # Third handle, complete.
    mock_socket.do_handshake.side_effect = None
    ret = TLSHandshakeState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_socket.do_handshake.assert_called_once()
    mock_socket.reset_mock()
    assert fsm.state is MQTTHandshakeConnectState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_tls_handshake_timeout(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mocker: MockerFixture,
    mock_select: Mock,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtts://testhost"), tls_context=mocker.Mock())
    fsm = FSM(env=env, init_state=TLSHandshakeState, error_state=ShutdownState)

    mock_timeout.exceeded.return_value = True
    TLSHandshakeState.enter(fsm, state_data, env, params)
    mock_timeout.mark.assert_not_called()
    ret = TLSHandshakeState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize(
    ("address", "user", "pwd"), [
        ("mqtt://testhost", None, None),
        ("mqtt://test_user:test_pass@testhost", "test_user", "test_pass"),
    ],
)
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connect_happy_path(
    address: str,
    user: str | None,
    pwd: str | None,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: Mock,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock,
    mocker: MockerFixture
) -> None:
    env.write_buffer = mocker.Mock()
    params = ConnectParams(
        address=Address(address),
        client_id="test_client",
        keepalive_interval=60,
        clean_start=True,
        will_topic="test_topic",
        will_payload=b"test_payload",
        will_qos=1,
        will_retain=True,
        will_properties=MQTTWillProps(),
        connect_properties=MQTTConnectProps(),
    )
    fsm = FSM(env=env, init_state=MQTTHandshakeConnectState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # Enter state.
    MQTTHandshakeConnectState.enter(fsm, state_data, env, params)
    env.write_buffer.clear.assert_called_once()
    env.write_buffer.extend.assert_called_once()
    data = env.write_buffer.extend.call_args[0][0]
    env.write_buffer.reset_mock()
    env.write_buffer = bytearray(data)
    packet = decode_packet(data)
    assert isinstance(packet, MQTTConnectPacket)
    assert packet.client_id == params.client_id
    assert packet.keep_alive == params.keepalive_interval
    assert packet.clean_start is params.clean_start
    assert packet.will_topic == params.will_topic
    assert packet.will_payload == params.will_payload
    assert packet.will_qos == params.will_qos
    assert packet.will_retain is params.will_retain
    assert packet.will_props == params.will_properties
    assert packet.properties == params.connect_properties
    assert packet.username == user
    assert packet.password == (pwd.encode() if pwd else None)
    assert fsm.state is MQTTHandshakeConnectState

    # First handle, blocked write.
    if params.address.use_tls:
        mock_socket.send.side_effect = ssl.SSLWantWriteError
    else:
        mock_socket.send.side_effect = BlockingIOError
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_socket.send.assert_called_once_with(env.write_buffer)
    mock_socket.reset_mock()
    assert fsm.state is MQTTHandshakeConnectState
    if max_wait:
        mock_select.select.assert_called_once_with(write=True, timeout=1)

    # Second handle, complete.
    mock_socket.send.side_effect = None
    mock_socket.send.return_value = len(data)
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_socket.send.assert_called_once_with(env.write_buffer)
    mock_socket.reset_mock()
    assert fsm.state is MQTTHandshakeConnAckState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connect_timeout(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnectState, error_state=ShutdownState)

    MQTTHandshakeConnectState.enter(fsm, state_data, env, params)
    mock_timeout.mark.assert_not_called()
    mock_timeout.exceeded.return_value = True
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connect_partial(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: Mock,
    mock_socket: Mock,
    mock_timeout: Mock,
    mocker: MockerFixture
) -> None:
    env.write_buffer = mocker.Mock()
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnectState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # One byte on the first write.
    MQTTHandshakeConnectState.enter(fsm, state_data, env, params)
    data = env.write_buffer.extend.call_args[0][0]
    env.write_buffer.reset_mock()
    env.write_buffer = bytearray(data)
    mock_socket.send.side_effect = None
    mock_socket.send.return_value = 1
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_socket.send.assert_called_once_with(env.write_buffer)
    mock_socket.reset_mock()
    assert env.write_buffer == bytearray(data[1:])
    assert fsm.state is MQTTHandshakeConnectState

    # The rest on the second write.
    mock_socket.send.return_value = len(data) - 1
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_socket.send.assert_called_once_with(env.write_buffer)
    mock_socket.reset_mock()
    assert env.write_buffer == bytearray()
    assert fsm.state is MQTTHandshakeConnAckState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connect_error(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnectState, error_state=ShutdownState)

    MQTTHandshakeConnectState.enter(fsm, state_data, env, params)
    mock_socket.send.return_value = 0
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connack_happy_path(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    decoder: Mock,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnAckState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # Enter state.
    MQTTHandshakeConnAckState.enter(fsm, state_data, env, params)
    assert fsm.state is MQTTHandshakeConnAckState

    # First handle, blocked read.
    decoder.decode.return_value = None
    mock_select.select.return_value = (True, False)
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(read=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    assert fsm.state is MQTTHandshakeConnAckState

    # Second handle, complete.
    connack = MQTTConnAckPacket(properties=MQTTConnAckProps(ServerKeepAlive=23))
    decoder.decode.return_value = connack
    mock_select.select.return_value = (False, False)
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert state_data.connack == connack
    assert state_data.keepalive.keepalive_interval == connack.properties.ServerKeepAlive
    assert fsm.state is ConnectedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connack_timeout(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnAckState, error_state=ShutdownState)

    MQTTHandshakeConnAckState.enter(fsm, state_data, env, params)
    mock_timeout.mark.assert_not_called()
    mock_timeout.exceeded.return_value = True
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connack_closed_socket(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    decoder: Mock,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnAckState, error_state=ShutdownState)

    MQTTHandshakeConnAckState.enter(fsm, state_data, env, params)
    decoder.decode.side_effect = ClosedSocketError("TEST")
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connack_unexpected(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    decoder: Mock,
    state_data: StateData,
    env: StateEnvironment
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnAckState, error_state=ShutdownState)

    MQTTHandshakeConnAckState.enter(fsm, state_data, env, params)
    decoder.decode.return_value = MQTTPublishPacket(topic="test/topic", payload=b"test_payload")
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_happy_path(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: Mock,
    mock_keepalive: Mock,
    mock_select: Mock,
    mock_socket: Mock,
    mock_read: Mock,
    mocker: MockerFixture
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)
    mock_keepalive.get_next_timeout.return_value = 1 if max_wait is None else max_wait

    # Enter state.
    ConnectedState.enter(fsm, state_data, env, params)
    mock_keepalive.mark_init.assert_called_once()
    callbacks.packet.assert_called_once_with(state_data.connack)
    callbacks.reset()
    assert state_data.open_called is True
    assert fsm.state is ConnectedState

    # First handle, nothing to read.
    mock_select.select.return_value = (False, False)
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(read=True, write=False, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    assert fsm.state is ConnectedState

    # Second handle, send data in write buffer.
    env.write_buffer.extend(b"test_data")
    mock_select.select.return_value = (False, True)
    mock_socket.send.return_value = len(env.write_buffer)
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(read=True, write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_socket.send.assert_called_once_with(env.write_buffer)
    mock_socket.reset_mock()
    assert len(env.write_buffer) == 0
    mock_keepalive.mark_send.assert_called_once()
    assert fsm.state is ConnectedState

    # Third handle, read data.
    mock_select.select.return_value = (True, False)
    mock_read.side_effect = [True, False]
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(read=True, write=False, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    assert mock_read.call_count == 2

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_keepalive(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_keepalive: Mock,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    ConnectedState.enter(fsm, state_data, env, params)
    callbacks.reset()

    # First handle, send a PINGREQ.
    mock_keepalive.should_send_ping.return_value = True
    mock_select.select.return_value = (False, True)
    mock_socket.send.return_value = len(PING)
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_keepalive.should_send_ping.assert_called_once()
    mock_keepalive.mark_ping.assert_called_once()
    mock_keepalive.reset_mock()
    mock_socket.send.assert_called_once_with(env.write_buffer)
    mock_socket.reset_mock()
    assert len(env.write_buffer) == 0

    # Second handle, keepalive timeout.
    mock_keepalive.should_send_ping.return_value = False
    mock_keepalive.should_close.return_value = True
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("block_exc", [ssl.SSLWantReadError, ssl.SSLWantWriteError, BlockingIOError])
@pytest.mark.parametrize("fatal_exc", [BrokenPipeError, ConnectionResetError])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_send_errors(
    max_wait: float | None,
    block_exc: type[Exception],
    fatal_exc: type[Exception],
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_read: Mock,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    ConnectedState.enter(fsm, state_data, env, params)
    callbacks.reset()

    env.write_buffer.extend(b"test_data")
    mock_read.return_value = False
    mock_select.select.return_value = (True, True)

    mock_socket.send.side_effect = block_exc("TEST")
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_read.assert_called_once()
    mock_read.reset_mock()
    assert fsm.state is ConnectedState

    mock_select.select.return_value = (True, True)
    mock_socket.send.side_effect = fatal_exc("TEST")
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_read.assert_not_called()
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_read_closed(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_read: Mock,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    ConnectedState.enter(fsm, state_data, env, params)
    callbacks.reset()

    mock_select.select.return_value = (True, False)
    mock_read.side_effect = ClosedSocketError("TEST")
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_read_mqtt_error(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_read: Mock,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    ConnectedState.enter(fsm, state_data, env, params)
    callbacks.reset()

    mock_select.select.return_value = (True, False)
    mock_read.side_effect = MQTTError("TEST", reason_code=MQTTReasonCode.ProtocolError)
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert state_data.disconnect_rc == MQTTReasonCode.ProtocolError
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


def test_states_connected_read_packet(
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    decoder: Mock,
    mock_keepalive: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    # Handle incomplete packet.
    decoder.decode.return_value = None
    ret = ConnectedState.read_packet(fsm, state_data, env, params)
    assert ret is False
    decoder.decode.assert_called_once()
    decoder.reset_mock()
    assert fsm.state is ConnectedState

    # Handle complete PUBLISH packet.
    pub_packet = MQTTPublishPacket(topic="test/topic", payload=b"test_payload")
    decoder.decode.return_value = pub_packet
    ret = ConnectedState.read_packet(fsm, state_data, env, params)
    assert ret is True
    decoder.decode.assert_called_once()
    decoder.reset_mock()
    callbacks.packet.assert_called_once_with(pub_packet)
    callbacks.reset()
    assert fsm.state is ConnectedState

    # Handle complete PINGRESP packet.
    pong_packet = MQTTPingRespPacket()
    decoder.decode.return_value = pong_packet
    ret = ConnectedState.read_packet(fsm, state_data, env, params)
    assert ret is True
    decoder.decode.assert_called_once()
    decoder.reset_mock()
    mock_keepalive.mark_pong.assert_called_once()
    callbacks.assert_not_called()
    assert fsm.state is ConnectedState

    # Handle complete PINGREQ packet.
    ping_packet = MQTTPingReqPacket()
    decoder.decode.return_value = ping_packet
    ret = ConnectedState.read_packet(fsm, state_data, env, params)
    assert ret is True
    decoder.decode.assert_called_once()
    decoder.reset_mock()
    assert env.write_buffer == PONG
    env.write_buffer.clear()
    callbacks.assert_not_called()
    assert fsm.state is ConnectedState

    # Handle complete DISCONNECT packet.
    dc_packet = MQTTDisconnectPacket()
    decoder.decode.return_value = dc_packet
    ret = ConnectedState.read_packet(fsm, state_data, env, params)
    assert ret is True
    decoder.decode.assert_called_once()
    decoder.reset_mock()
    callbacks.assert_not_called()
    assert fsm.state is ClosingState


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_reconnect_wait_happy_path(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mocker: MockerFixture,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), reconnect_delay=5)
    fsm = FSM(env=env, init_state=ReconnectWaitState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait
    mock_cond = mocker.MagicMock(spec=Condition)
    mock_cond.wait.return_value = True
    mock_cond.__enter__.return_value = mock_cond
    fsm.cond = mock_cond

    ReconnectWaitState.enter(fsm, state_data, env, params)
    assert mock_timeout.interval == params.reconnect_delay
    mock_timeout.mark.assert_called_once()
    mock_timeout.reset_mock()

    # First handle, waiting.
    ret = ReconnectWaitState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_timeout.exceeded.assert_called_once()
    mock_timeout.reset_mock()
    if max_wait is None:
        mock_cond.wait.assert_called_once_with(mock_timeout.get_timeout.return_value)
        mock_cond.wait.reset_mock()
    else:
        mock_cond.wait.assert_not_called()
    assert fsm.state is ReconnectWaitState

    # Second handle, transition.
    mock_timeout.exceeded.return_value = True
    ret = ReconnectWaitState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_timeout.exceeded.assert_called_once()
    mock_cond.wait.assert_not_called()
    assert fsm.state is ConnectingState

    callbacks.assert_not_called()


def test_states_reconnect_wait_race(
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mocker: MockerFixture
) -> None:
    """Check for a race condition between state changes and waiting for state changes."""
    params = ConnectParams(address=Address("mqtt://testhost"), reconnect_delay=5)
    fsm = FSM(env=env, init_state=ReconnectWaitState, error_state=ShutdownState)
    mock_cond = mocker.MagicMock(spec=Condition)
    mock_cond.wait.return_value = True
    mock_cond.__enter__.return_value = mock_cond
    fsm.cond = mock_cond

    ReconnectWaitState.enter(fsm, state_data, env, params)
    fsm.change_state(ConnectingState)
    ret = ReconnectWaitState.handle(fsm, state_data, env, params, True)
    assert ret is True

    callbacks.assert_not_called()


@pytest.mark.parametrize("address", ["mqtt://testhost", "mqtts://testhost"])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_closing_happy_path(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address(address), connect_timeout=5)
    fsm = FSM(env=env, init_state=ClosingState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # Start from ConnectedState to test transition through ClosingState to ClosedState.
    fsm.previous_state = ConnectedState
    ClosingState.enter(fsm, state_data, env, params)
    assert mock_timeout.interval == params.connect_timeout
    mock_timeout.mark.assert_called_once()
    mock_timeout.reset_mock()
    assert state_data.disconnect_rc == MQTTReasonCode.NormalDisconnection
    if not params.address.use_tls:
        mock_socket.shutdown.assert_called_once_with(socket.SHUT_RD)
    else:
        mock_socket.shutdown.assert_not_called()
    mock_socket.reset_mock()
    assert fsm.state is ClosingState

    # First handle, write all.
    mock_select.select.return_value = (False, True)
    env.write_buffer.extend(b"test_data")
    mock_socket.send.return_value = len(env.write_buffer)
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_socket.send.assert_called_once_with(env.write_buffer)
    mock_socket.reset_mock()
    mock_timeout.exceeded.assert_called_once()
    mock_timeout.reset_mock()
    assert fsm.state is ClosingState

    # Second handle, done writing, transition to ClosedState.
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_timeout.exceeded.assert_called_once()
    mock_timeout.reset_mock()
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


# Try all states which can transition to ClosingState, except ConnectedState.
@pytest.mark.parametrize("prev_state", [
    ConnectingState,
    TLSHandshakeState,
    MQTTHandshakeConnectState,
    MQTTHandshakeConnAckState,
    ReconnectWaitState,
])
def test_states_closing_skip(
    prev_state: Any,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=ClosingState, error_state=ShutdownState)

    fsm.previous_state = prev_state
    ClosingState.enter(fsm, state_data, env, params)
    mock_socket.shutdown.assert_not_called()
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_closing_timeout(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=ClosingState, error_state=ShutdownState)

    fsm.previous_state = ConnectedState
    ClosingState.enter(fsm, state_data, env, params)
    mock_timeout.exceeded.return_value = True
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


def test_states_closing_shutdown_error(
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=ClosingState, error_state=ShutdownState)

    fsm.previous_state = ConnectedState
    mock_socket.shutdown.side_effect = OSError("TEST")
    ClosingState.enter(fsm, state_data, env, params)
    mock_socket.shutdown.assert_called_once()
    assert fsm.state is ClosingState

    callbacks.assert_not_called()


@pytest.mark.parametrize("exc", [ssl.SSLWantReadError, ssl.SSLWantWriteError, BlockingIOError])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_closing_send_error(
    max_wait: float | None,
    exc: type[Exception],
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=ClosingState, error_state=ShutdownState)

    fsm.previous_state = ConnectedState
    ClosingState.enter(fsm, state_data, env, params)

    env.write_buffer.extend(b"test_data")
    mock_select.select.return_value = (False, True)
    mock_socket.send.side_effect = exc("TEST")
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    assert fsm.state is ClosingState

    mock_socket.send.side_effect = BrokenPipeError("TEST")
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("open_called", [True, False])
@pytest.mark.parametrize("rc", [None, MQTTReasonCode.NormalDisconnection])
@pytest.mark.parametrize("reconn_delay", [0, 1])
def test_states_closed_happy_path(
    open_called: bool,
    rc: MQTTReasonCode | None,
    reconn_delay: int,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    decoder: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), reconnect_delay=reconn_delay)
    fsm = FSM(env=env, init_state=ClosedState, error_state=ShutdownState)

    fsm.requested_state = ConnectingState
    state_data.disconnect_rc = rc
    state_data.open_called = open_called
    ClosedState.enter(fsm, state_data, env, params)
    if open_called:
        if rc is not None:
            mock_socket.send.assert_called_once_with(MQTTDisconnectPacket().encode())
    mock_socket.shutdown.assert_called_once_with(socket.SHUT_RDWR)
    mock_socket.close.assert_called_once()
    decoder.reset.assert_called_once()
    # We will assert that the write buffer is empty in another test.
    if reconn_delay > 0 and open_called:
        assert fsm.state is ReconnectWaitState
    else:
        assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("exc", [ssl.SSLWantReadError, ssl.SSLWantWriteError, BlockingIOError, OSError])
def test_states_closed_errors(
    exc: type[Exception],
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ClosedState, error_state=ShutdownState)

    fsm.previous_state = ClosingState
    state_data.open_called = True
    state_data.disconnect_rc = MQTTReasonCode.NormalDisconnection
    mock_socket.send.side_effect = exc("TEST")
    mock_socket.shutdown.side_effect = OSError("TEST")
    mock_socket.close.side_effect = OSError("TEST")
    ClosedState.enter(fsm, state_data, env, params)
    mock_socket.send.assert_called_once_with(MQTTDisconnectPacket().encode())
    mock_socket.shutdown.assert_called_once_with(socket.SHUT_RDWR)
    mock_socket.close.assert_called_once()
    assert state_data.open_called is False
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


def test_states_closed_write_buffer_not_empty(
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ClosedState, error_state=ShutdownState)

    fsm.previous_state = ConnectingState
    env.write_buffer.extend(b"test_data")
    ClosedState.enter(fsm, state_data, env, params)
    mock_socket.send.assert_not_called()
    assert len(env.write_buffer) == 0
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("close_exc", [OSError, None])
@pytest.mark.parametrize("open_called", [True, False])
def test_states_shutdown_enter(
    open_called: bool,
    close_exc: type[Exception] | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: Mock,
    decoder: Mock,
    mocker: MockerFixture,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ShutdownState, error_state=ShutdownState)

    env.write_buffer = mocker.Mock()
    state_data.open_called = open_called
    mock_socket.close.side_effect = close_exc("TEST") if close_exc else None
    ShutdownState.enter(fsm, state_data, env, params)
    if open_called:
        pass
    mock_socket.close.assert_called_once()
    decoder.reset.assert_called_once()
    env.write_buffer.clear.assert_called_once()
    assert fsm.state is ShutdownState

    callbacks.assert_not_called()



from __future__ import annotations

import socket
import ssl
from typing import cast, Final

from .types import ConnectParams, StateData, StateEnvironment, ReceivablePacketT
from .fsm import FSM, FSMState
from .decoder import ClosedSocketError
from ..error import MQTTError
from ..logger import get_logger
from ..mqtt_spec import MQTTPacketType, MQTTReasonCode
from ..packet import MQTTConnectPacket, MQTTConnAckPacket, MQTTDisconnectPacket, PING, PONG
from ..platform import AF_UNIX, HAS_AF_UNIX

logger: Final = get_logger("connection.states")


def _get_socket(family: socket.AddressFamily) -> socket.socket:
    """Get a socket object.

    This is patched in tests to use a mock or loopback socket."""
    return socket.socket(family, socket.SOCK_STREAM)


class ConnectingState(FSMState):
    """Connecting to the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.keepalive.keepalive_interval = params.keepalive_interval
        state_data.timeout.interval = params.connect_timeout
        state_data.timeout.mark()
        state_data.connack = None
        state_data.disconnect_rc = None
        state_data.sock = _get_socket(params.address.family)
        if params.address.family in (socket.AF_INET, socket.AF_INET6):
            state_data.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, params.tcp_nodelay)
        state_data.decoder.reset()
        state_data.open_called = False
        with fsm.selector:
            fsm.selector.change_sock(state_data.sock)

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("Connect timeout")
            fsm.change_state(ClosedState)
            return True

        try:
            address = params.address
            if HAS_AF_UNIX and address.family == AF_UNIX:
                state_data.sock.connect(address.host)
            else:
                state_data.sock.connect((address.host, address.port))
        except ConnectionError as exc:
            logger.error("Failed to connect to broker: %s", exc)
            fsm.change_state(ClosedState)
            return True
        except (BlockingIOError, OSError):
            pass  # Either in progress or already connected, select will reveal which.

        with fsm.selector:
            timeout = state_data.timeout.get_timeout(max_wait)
            _, writable = fsm.selector.select(write=True, timeout=timeout)
        if writable:
            logger.debug("Socket connected to broker")
            if params.address.use_tls:
                fsm.change_state(TLSHandshakeState)
            else:
                fsm.change_state(MQTTHandshakeConnectState)
            state_data.sock.setblocking(False)
            return True
        return False


class TLSHandshakeState(FSMState):
    """Performing TLS handshake with the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        tls_context = params.tls_context if params.tls_context else ssl.create_default_context()
        state_data.sock = tls_context.wrap_socket(
            state_data.sock,
            server_hostname=params.tls_hostname if params.tls_hostname else params.address.host,
            do_handshake_on_connect=False,
        )
        with fsm.selector:
            fsm.selector.change_sock(state_data.sock)

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("TLS handshake timeout")
            fsm.change_state(ClosedState)
            return True

        timeout = state_data.timeout.get_timeout(max_wait)
        sock = cast(ssl.SSLSocket, state_data.sock)
        try:
            logger.debug("trying TLS handshake")
            sock.do_handshake()
            fsm.change_state(MQTTHandshakeConnectState)
            return True
        except ssl.SSLWantReadError:
            logger.debug("TLS handshake wants read")
            with fsm.selector:
                fsm.selector.select(read=True, timeout=timeout)
        except ssl.SSLWantWriteError:
            logger.debug("TLS handshake wants write")
            with fsm.selector:
                fsm.selector.select(write=True, timeout=timeout)
        return False


class MQTTHandshakeConnectState(FSMState):
    """Sending MQTT CONNECT packet to the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        connect_packet = MQTTConnectPacket(
            client_id=params.client_id,
            protocol_version=params.protocol_version,
            clean_start=params.clean_start,
            keep_alive=params.keepalive_interval,
            properties=params.connect_properties,
            will_topic=params.will_topic,
            will_payload=params.will_payload,
            will_qos=params.will_qos,
            will_retain=params.will_retain,
            will_props=params.will_properties,
            username=params.address.username,
            password=params.address.password.encode() if params.address.password else None,
        )
        logger.debug("---> %s", connect_packet)
        with fsm.selector:
            env.write_buffer.clear()
            env.write_buffer.extend(connect_packet.encode())

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("MQTT CONNECT keepalive timeout")
            fsm.change_state(ClosedState)
            return True

        try:
            with fsm.selector:
                num_sent = state_data.sock.send(env.write_buffer)
                if num_sent == 0:
                    logger.error("MQTT CONNECT send returned 0 bytes, closing connection")
                    fsm.change_state(ClosedState)
                    return True
                if num_sent < len(env.write_buffer):
                    # Not all data was sent, wait for writable again.
                    logger.debug("Not all CONNECT data was sent, waiting for writable again: wrote: %d", num_sent)
                    del env.write_buffer[:num_sent]
                    return False
                env.write_buffer.clear()
            fsm.change_state(MQTTHandshakeConnAckState)
            return True
        except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
            # The write was blocked, wait for the socket to be writable.
            if max_wait is None or max_wait > 0.0:
                with fsm.selector:
                    timeout = state_data.timeout.get_timeout(max_wait)
                    fsm.selector.select(write=True, timeout=timeout)
        return False


class MQTTHandshakeConnAckState(FSMState):
    """Receiving MQTT CONNACK packet from the broker."""
    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("MQTT CONNACK keepalive timeout")
            fsm.change_state(ClosedState)
            return True

        want_read = False
        try:
            packet = state_data.decoder.decode(state_data.sock)
            if packet is None:
                want_read = True
        except ClosedSocketError:
            logger.exception("Socket was closed")
            fsm.change_state(ClosedState)
            return True

        if want_read:
            # Incomplete packet, wait for more data.
            with fsm.selector:
                timeout = state_data.timeout.get_timeout(max_wait)
                fsm.selector.select(read=True, timeout=timeout)
            return False

        if packet is not None and packet.packet_type == MQTTPacketType.CONNACK:
            packet = cast(MQTTConnAckPacket, packet)
            state_data.connack = packet
            if packet.properties.ServerKeepAlive is not None:
                state_data.keepalive.keepalive_interval = packet.properties.ServerKeepAlive
            fsm.change_state(ConnectedState)
            return True
        pt = packet.packet_type.name if packet is not None else "None"
        logger.error("Unexpected '%s' packet while waiting for CONNACK", pt)
        fsm.change_state(ClosedState)
        return True


class ConnectedState(FSMState):
    """Connected to the broker. Full duplex messaging in this state."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.keepalive.mark_init()
        with fsm.selector:
            env.write_buffer.clear()
        assert state_data.connack is not None
        env.packet_callback(state_data.connack)
        state_data.open_called = True

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.keepalive.should_close():
            logger.error("Keepalive timeout, closing socket")
            fsm.change_state(ClosedState)
            return True

        if state_data.keepalive.should_send_ping():
            logger.debug("---> PING")
            with fsm.selector:
                env.write_buffer.extend(PING)
            state_data.keepalive.mark_ping()

        timeout = state_data.keepalive.get_next_timeout(max_wait)
        with fsm.selector:
            write_check = bool(env.write_buffer)
            readable, writable = fsm.selector.select(read=True, write=write_check, timeout=timeout)

        if writable:
            try:
                with fsm.selector:
                    sent = state_data.sock.send(env.write_buffer)
                    del env.write_buffer[:sent]
                state_data.keepalive.mark_send()
            except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
                pass
            except (BrokenPipeError, ConnectionResetError) as exc:
                logger.error("MQTT connection was closed: %s", exc)
                fsm.change_state(ClosedState)
                return True

        if readable:
            want_read = True
            try:
                while want_read:  # Read all available packets.
                    want_read = cls.read_packet(fsm, state_data, env, params)
            except ClosedSocketError:
                logger.debug("Connection closed")
                fsm.change_state(ClosedState)
                return True
            except MQTTError as exc:
                logger.error("There was a problem with data from broker, closing connection: %s", exc)
                state_data.disconnect_rc = exc.reason_code
                fsm.change_state(ClosedState)
                return True

        return False

    @classmethod
    def read_packet(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> bool:
        """Called by the underlying SocketWrapper when the socket is ready to read.

        Incrementally reads and decodes a packet from the socket.
        Complete packets are passed up to the read callback.

        Returns False if an incomplete packet was read, True if a complete packet was read."""
        packet = state_data.decoder.decode(state_data.sock)
        if packet is None:
            # No complete packet available yet.
            return False

        # Ping requests and responses are handled at this layer.
        if packet.packet_type == MQTTPacketType.PINGRESP:
            logger.debug("<--- PONG")
            state_data.keepalive.mark_pong()
        elif packet.packet_type == MQTTPacketType.PINGREQ:
            logger.debug("<--- PING PONG --->")
            with fsm.selector:
                env.write_buffer.extend(PONG)
        elif packet.packet_type == MQTTPacketType.DISCONNECT:
            logger.debug("<--- %s", packet)
            logger.info("Broker sent DISCONNECT, closing connection")
            fsm.change_state(ClosingState)
            return True
        else:
            # All other packets are passed to the read callback.
            try:
                # To cast here, we must handle the exceptional case at runtime.
                env.packet_callback(cast(ReceivablePacketT, packet))
            except KeyError as exc:
                raise MQTTError("Unexpected packet type", reason_code=MQTTReasonCode.ProtocolError) from exc
        return True


class ReconnectWaitState(FSMState):
    """Waiting to reconnect to the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.timeout.interval = params.reconnect_delay
        state_data.timeout.mark()

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        # Here we repurpose the keepalive timer to wait for a reconnect.
        if state_data.timeout.exceeded():
            logger.debug("Reconnecting")
            fsm.change_state(ConnectingState)
            return True
        with fsm.cond:
            if fsm.state is not ReconnectWaitState:
                # The state has changed, don't wait.
                return True
            if max_wait is None or max_wait > 0.0:
                timeout = state_data.timeout.get_timeout(max_wait)
                fsm.cond.wait(timeout)
        return False


class ClosingState(FSMState):
    """Gracefully closing the connection.

    The socket will be shutdown for reading, but existing buffers will be flushed."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        if fsm.previous_state == ConnectedState:
            logger.debug("Gracefully closing connection")
            if state_data.disconnect_rc is None:
                state_data.disconnect_rc = MQTTReasonCode.NormalDisconnection
        else:
            logger.debug("Skipping ClosingState")
            fsm.change_state(ClosedState)
            return

        state_data.timeout.interval = params.connect_timeout
        state_data.timeout.mark()

        # Shutdown only the read side of the socket, so we can still send data.
        # We don't want to shutdown the socket if we're using TLS, it would cause the protocol to fail.
        if not params.address.use_tls:
            try:
                state_data.sock.shutdown(socket.SHUT_RD)
            except OSError as exc:
                logger.debug("Error while shutting down socket reading: %s", exc)

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        # Wait for the socket to be writable.
        if state_data.timeout.exceeded():
            logger.error("ClosingState timed out")
            fsm.change_state(ClosedState)
            return True

        with fsm.selector:
            if not env.write_buffer:
                logger.debug("No more data to send, connection closed")
                fsm.change_state(ClosedState)
                return True
            timeout = state_data.timeout.get_timeout(max_wait)
            _, writable = fsm.selector.select(write=True, timeout=timeout)

            if writable:
                try:
                    sent = state_data.sock.send(env.write_buffer)
                    del env.write_buffer[:sent]
                    state_data.keepalive.mark_send()
                except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
                    pass
                except BrokenPipeError:
                    logger.error("Socket lost while closing")
                    fsm.change_state(ClosedState)
                    return True
        return False


class ClosedState(FSMState):
    """Connection is closed.

    All buffers will be flushed and the socket will be closed after conditionally sending a DISCONNECT."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        if state_data.open_called:
            state_data.open_called = False
            with fsm.selector:
                if state_data.disconnect_rc is not None and not env.write_buffer:
                    disconnect_packet = MQTTDisconnectPacket(reason_code=state_data.disconnect_rc)
                    # Try to send a DISCONNECT packet, but no problem if we can't.
                    try:
                        state_data.sock.send(disconnect_packet.encode())
                        logger.debug("---> %s", disconnect_packet)
                    except OSError:
                        logger.debug("Failed to send DISCONNECT packet")
            if params.reconnect_delay > 0 and fsm.requested_state == ConnectingState:
                fsm.change_state(ReconnectWaitState)
        try:
            state_data.sock.shutdown(socket.SHUT_RDWR)
        except OSError as exc:
            logger.debug("Error while shutting down socket: %s", exc)
        try:
            state_data.sock.close()
        except OSError as exc:
            logger.debug("Error while closing socket: %s", exc)
        state_data.decoder.reset()
        with fsm.selector:
            env.write_buffer.clear()


class ShutdownState(FSMState):
    """The final state.

    All buffers will be flushed and the socket will be closed immediately on entry."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        # We can enter this state from any other state.
        # Free up as many resources as possible.
        try:
            state_data.sock.close()
        except OSError as exc:
            logger.debug("Error while closing socket: %s", exc)
        state_data.decoder.reset()
        with fsm.selector:
            env.write_buffer.clear()
            fsm.selector.close()


# Hook up transitions.
ConnectingState.transitions_to = (ClosingState, ClosedState, ShutdownState, TLSHandshakeState, MQTTHandshakeConnectState)
ConnectingState.can_request_from = (ClosedState, ReconnectWaitState)

TLSHandshakeState.transitions_to = (ClosingState, ClosedState, ShutdownState, MQTTHandshakeConnectState)

MQTTHandshakeConnectState.transitions_to = (ClosingState, ClosedState, ShutdownState, MQTTHandshakeConnAckState)

MQTTHandshakeConnAckState.transitions_to = (ClosingState, ClosedState, ShutdownState, ConnectedState)

ConnectedState.transitions_to = (ClosingState, ClosedState, ShutdownState)

ClosingState.transitions_to = (ClosedState, ShutdownState)
ClosingState.can_request_from = (
    ConnectingState,
    TLSHandshakeState,
    MQTTHandshakeConnectState,
    MQTTHandshakeConnAckState,
    ConnectedState,
    ReconnectWaitState,
)

ClosedState.transitions_to = (ConnectingState, ShutdownState, ReconnectWaitState)

ReconnectWaitState.transitions_to = (ClosingState, ClosedState, ShutdownState, ConnectingState)

ShutdownState.can_request_from = (
    ConnectingState,
    TLSHandshakeState,
    MQTTHandshakeConnectState,
    MQTTHandshakeConnAckState,
    ConnectedState,
    ReconnectWaitState,
    ClosingState,
    ClosedState,
)



import nox
import sys

nox.options.default_venv_backend = "uv|virtualenv"
py_versions = ("3.10", "3.11", "3.12", "3.13")


@nox.session(python=py_versions)
def tests(session):
    session.install(".")
    session.install("--group", "dev")
    session.run("ruff", "check")
    session.run("typos")
    session.run("mypy")
    if sys.platform != "win32":
        session.run("complexipy", "-d", "low", "ohmqtt", "examples", "tests")
    else:
        session.log("Skipping complexipy check on Windows: https://github.com/rohaquinlop/complexipy/issues/67")
    session.run("pytest", "--cov-report=term-missing")
